;; UNIX-style Brace Expansion
;; (A solution to <http://www.rosettacode.org/wiki/Brace_expansion>.)
;;
;; Edwin Watkeys, Thunk NYC Corp.
;; November 26, 2018
;;

(define (build-tree chs)
  (let loop ((chs chs) (accum '()) (options '()))
    (if (null? chs) (error "Missing closing bracket")
	(let ((ch (car chs)))
	  (cond ((char=? #\} ch)
		 (values (reverse (cons (reverse accum) options))
			 (cdr chs)))
		((char=? #\, ch)
		 (loop (cdr chs) '() (cons (reverse accum) options)))
		((char=? #\{ ch)
		 (let-values (((subtree rest-chs) (build-tree (cdr chs))))
		   (loop rest-chs (cons subtree accum) options)))
		(else
		 (loop (cdr chs) (cons ch accum) options)))))))

(define (build-root chs)
  (let loop ((chs chs) (accum '()))
    (if (null? chs) (reverse accum)
	(let ((ch (car chs)))
	  (if (not (char=? #\{ ch))
	      (loop (cdr chs) (cons ch accum))
	      (let-values (((subtree rest-chs) (build-tree (cdr chs))))
		(loop rest-chs (cons subtree accum))))))))

(define (mappend f xs)
  (reduce-right (lambda (x a) (append (f x) a)) '() xs))

(define (expand-subtrees ts) (mappend expand-tree ts))

(define (cross-choices choices expansion)
  (mappend (lambda (c) (map (lambda (el) (append c el)) expansion)) choices))

(define (expand-tree tree)
  (fold (lambda (node accum)
	  (if (list? node)
	      (cross-choices (expand-subtrees node) accum)
	      (map (lambda (el) (cons node el)) accum)))
	'(()) tree))

;;> Expand curly braces in \var(s) a la Bourne shell into a list of strings.

(define (expand-braces s)
  (map reverse-list->string (expand-tree (build-root (string->list s)))))
